"""Hierarchical Clustering estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from skfolio.cluster._hierarchical import HierarchicalClustering, LinkageMethod

__all__ = ["LinkageMethod", "HierarchicalClustering"]
